# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tech.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'tech.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import process_stack_pb2 as process__stack__pb2
import process_parasitics_pb2 as process__parasitics__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\ntech.proto\x12\tkpex.tech\x1a\x13process_stack.proto\x1a\x18process_parasitics.proto\"\xef\x01\n\nTechnology\x12\x0c\n\x04name\x18\x01 \x01(\t\x12$\n\x06layers\x18\x65 \x03(\x0b\x32\x14.kpex.tech.LayerInfo\x12\x39\n\x13lvs_computed_layers\x18x \x03(\x0b\x32\x1c.kpex.tech.ComputedLayerInfo\x12\x33\n\rprocess_stack\x18\x8c\x01 \x01(\x0b\x32\x1b.kpex.tech.ProcessStackInfo\x12=\n\x12process_parasitics\x18\xc8\x01 \x01(\x0b\x32 .kpex.tech.ProcessParasiticsInfo\"W\n\tLayerInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x0b \x01(\t\x12\x11\n\tgds_layer\x18\x15 \x01(\r\x12\x14\n\x0cgds_datatype\x18\x1f \x01(\r\"\xf0\x01\n\x11\x43omputedLayerInfo\x12/\n\x04kind\x18\n \x01(\x0e\x32!.kpex.tech.ComputedLayerInfo.Kind\x12(\n\nlayer_info\x18\x14 \x01(\x0b\x32\x14.kpex.tech.LayerInfo\x12\x1b\n\x13original_layer_name\x18\x1e \x01(\t\"c\n\x04Kind\x12\x14\n\x10KIND_UNSPECIFIED\x10\x00\x12\x10\n\x0cKIND_REGULAR\x10\x01\x12\x19\n\x15KIND_DEVICE_CAPACITOR\x10\x02\x12\x18\n\x14KIND_DEVICE_RESISTOR\x10\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tech_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_TECHNOLOGY']._serialized_start=73
  _globals['_TECHNOLOGY']._serialized_end=312
  _globals['_LAYERINFO']._serialized_start=314
  _globals['_LAYERINFO']._serialized_end=401
  _globals['_COMPUTEDLAYERINFO']._serialized_start=404
  _globals['_COMPUTEDLAYERINFO']._serialized_end=644
  _globals['_COMPUTEDLAYERINFO_KIND']._serialized_start=545
  _globals['_COMPUTEDLAYERINFO_KIND']._serialized_end=644
# @@protoc_insertion_point(module_scope)



#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
from enum import StrEnum
import time
from typing import *

import os
import subprocess

from ..log import (
    info,
    # warning,
    rule,
    subproc,
)
from ..version import __version__


class MagicPEXMode(StrEnum):
    CC = "CC"
    RC = "RC"
    R = "R"
    DEFAULT = "CC"


def prepare_magic_script(gds_path: str,
                         cell_name: str,
                         run_dir_path: str,
                         script_path: str,
                         output_netlist_path: str,
                         pex_mode: MagicPEXMode,
                         c_threshold: float,
                         r_threshold: float,
                         tolerance: float,
                         halo: Optional[float]):
    gds_path = os.path.abspath(gds_path)
    run_dir_path = os.path.abspath(run_dir_path)
    output_netlist_path = os.path.abspath(output_netlist_path)

    halo_scale = 200.0
    halo_decl = '' if halo is None else f"\nextract halo {round(halo * halo_scale)}"

    script: str = ""
    match pex_mode:
        case MagicPEXMode.CC:
            script = f"""# Generated by kpex {__version__}
crashbackups stop
drc off
gds read {gds_path}
load {cell_name}
select top cell
flatten {cell_name}_flat
load {cell_name}_flat
cellname delete {cell_name} -noprompt
cellname rename {cell_name}_flat {cell_name}
select top cell
extract path {run_dir_path}{halo_decl}
extract all
ext2spice cthresh {c_threshold}
ext2spice format ngspice
ext2spice -p {run_dir_path} -o {output_netlist_path}
quit -noprompt"""
        case MagicPEXMode.RC:
            script = f"""# Generated by kpex {__version__}
crashbackups stop
drc off
gds read {gds_path}
load {cell_name}
select top cell
flatten {cell_name}_flat
load {cell_name}_flat
cellname delete {cell_name} -noprompt
cellname rename {cell_name}_flat {cell_name}
select top cell
extract path {run_dir_path}{halo_decl}
extract do resistance
extract all
ext2sim labels on
ext2sim
extresist tolerance {tolerance}
extresist all
ext2spice cthresh {c_threshold}
ext2spice rthresh {r_threshold}
ext2spice extresist on
ext2spice format ngspice
ext2spice -p {run_dir_path} -o {output_netlist_path}
quit -noprompt
"""
        case MagicPEXMode.R:
            script = f"""# Generated by kpex {__version__}
crashbackups stop
drc off
gds read {gds_path}
load {cell_name}
select top cell
flatten {cell_name}_flat
load {cell_name}_flat
cellname delete {cell_name} -noprompt
cellname rename {cell_name}_flat {cell_name}
select top cell
extract path {run_dir_path}{halo_decl}
extract do resistance
extract no capacitance
extract no coupling
extract all
ext2sim labels on
ext2sim
extresist tolerance {tolerance}
extresist all
ext2spice rthresh {r_threshold}
ext2spice extresist on
ext2spice format ngspice
ext2spice -p {run_dir_path} -o {output_netlist_path}
quit -noprompt
"""

    with open(script_path, 'w') as f:
        f.write(script)

def run_magic(exe_path: str,
              magicrc_path: str,
              script_path: str,
              log_path: str):
    args = [
        exe_path,
        '-dnull',                      #
        '-noconsole',                  #
        '-rcfile',                     #
        magicrc_path,                  #
        script_path,                   # TCL script
    ]

    info('Calling MAGIC')
    subproc(f"{' '.join(args)}, output file: {log_path}")

    rule('MAGIC Output')

    start = time.time()

    proc = subprocess.Popen(args,
                            stdin=subprocess.DEVNULL,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            universal_newlines=True,
                            text=True)
    with open(log_path, 'w') as f:
        while True:
            line = proc.stdout.readline()
            if not line:
                break
            subproc(line[:-1])  # remove newline
            f.writelines([line])
    proc.wait()

    duration = time.time() - start

    rule()

    if proc.returncode == 0:
        info(f"MAGIC succeeded after {'%.4g' % duration}s")
    else:
        raise Exception(f"MAGIC failed with status code {proc.returncode} after {'%.4g' % duration}s, "
                        f"see log file: {log_path}")




#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
import gzip
import io
import os.path
import shutil
import sys
from typing import *
import xml.etree.ElementTree as ET

import klayout.rdb as rdb
from ..log import (
    LogLevel,
    set_log_level,
    register_additional_handler,
    deregister_additional_handler,
    # console,
    # debug,
    info,
    warning,
    subproc,
    error,
    rule
)


def parse_category_path(category_path: str) -> List[str]:
    within_escaped = False
    within_backslash = False
    current_word = ''
    path_list = []
    for c in category_path:
        match c:
            case '.':
                if within_backslash:
                    current_word += c
                    within_backslash = False
                elif within_escaped:
                    current_word += c
                else:
                    path_list.append(current_word)
                    current_word = ''
            case '\\':
                if within_backslash:
                    current_word += c
                    within_backslash = False
                else:
                    within_backslash = True
            case '\'':
                if within_backslash:
                    current_word += c
                    within_backslash = False
                else:
                    within_escaped = not within_escaped
            case _:
                current_word += c
    if len(current_word) >= 1:
        path_list.append(current_word)
    return path_list

def repair_rdb_xml(xml_file: io.IOBase, new_xml_path: str):
    et = ET.parse(xml_file)
    root = et.getroot()

    categories: Set[str] = set(
        [e.text for e in root.findall('./items/item/category')]
    )
    category_paths = [parse_category_path(c) for c in categories]
    category_paths.sort()
    # print(category_paths)
    for p in category_paths:
        elem = root
        for c in p:
            elemcats = elem.find("./categories")
            subelem = elemcats.find("./category/name[.='{0}']/..".format(c))
            if subelem is None:
                warning(f"In category path {p}, can't find element for component {c}")
                new_category = ET.SubElement(elemcats, "category")
                new_cname = ET.SubElement(new_category, "name")
                new_cname.text = c
                ET.SubElement(new_category, 'description')
                ET.SubElement(new_category, 'categories')
                elem = new_category
            else:
                elem = subelem

    et.write(new_xml_path)


def repair_rdb(rdb_path: str):
    rdb_file: io.IOBase
    suffix = os.path.splitext(rdb_path)[-1]
    new_xml_path = rdb_path + '.repair.xml'

    if suffix == '.gz':
        with gzip.open(rdb_path, 'r') as f:
            repair_rdb_xml(f, new_xml_path)
    else:
        with open(rdb_path, 'r') as f:
            repair_rdb_xml(f, new_xml_path)

    report = rdb.ReportDatabase('')
    try:
        report.load(new_xml_path)
        info(f"Succeeded in repairing broken marker database {rdb_path} under {new_xml_path}")

    except Exception as e:
        error(f"Failed to repair broken marker database {rdb_path} due to exception: {e}")


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(f"Usage: {sys.argv[0]} file.rdb.gz")
        sys.exit(1)

    repair_rdb(sys.argv[1])



#! /usr/bin/env python3
#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#

from enum import StrEnum
from dataclasses import dataclass
from typing import *


@dataclass
class Port:
    net: str
    x_bot: int
    y_bot: int
    x_top: int
    y_top: int
    layer: str


@dataclass
class Node:
    net: str
    int_r: int
    fin_c: int
    x_bot: int
    y_bot: int
    layer: str


class DeviceType(StrEnum):
    FET = "fet"
    MOSFET = "mosfet"
    ASSYMETRIC = "asymmetric"
    BJT = "bjt"
    DEVRES = "devres"
    DEVCAP = "devcap"
    DEVCAPREV = "devcaprev"
    VSOURCE = "vsource"
    DIODE = "diode"
    PDIODE = "pdiode"
    NDIODE = "ndiode"
    SUBCKT = "subckt"
    RSUBCKT = "rsubckt"
    MSUBCKT = "msubckt"
    CSUBCKT = "csubckt"


@dataclass
class Device:
    device_type: DeviceType
    model: str
    x_bot: int
    y_bot: int
    x_top: int
    y_top: int


@dataclass
class ExtData:
    ports: List[Port]
    nodes: List[Node]
    devices: List[Device]


@dataclass
class ResNode:
    name: str
    int_r: int
    fin_c: int
    x_bot: int
    y_bot: int


@dataclass
class Resistor:
    node1: str
    node2: str
    value_ohm: float


@dataclass
class ResExtData:
    rnodes: List[ResNode]
    resistors: List[Resistor]

    def rnodes_by_name(self, name: str) -> List[ResNode]:
        return [n for n in self.rnodes if n.name == name]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: process_stack.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'process_stack.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13process_stack.proto\x12\tkpex.tech\"\xb5\x0f\n\x10ProcessStackInfo\x12\x35\n\x06layers\x18\x64 \x03(\x0b\x32%.kpex.tech.ProcessStackInfo.LayerInfo\x1a?\n\x07\x43ontact\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0bmetal_above\x18\n \x01(\t\x12\x11\n\tthickness\x18\x14 \x01(\x01\x1a\x46\n\x0eSubstrateLayer\x12\x0e\n\x06height\x18\x01 \x01(\x01\x12\x11\n\tthickness\x18\x02 \x01(\x01\x12\x11\n\treference\x18\x1e \x01(\t\x1ak\n\nNWellLayer\x12\x0e\n\x06height\x18\x01 \x01(\x01\x12\x11\n\treference\x18\x1e \x01(\t\x12:\n\rcontact_above\x18( \x01(\x0b\x32#.kpex.tech.ProcessStackInfo.Contact\x1ao\n\x0e\x44iffusionLayer\x12\x0e\n\x06height\x18\x01 \x01(\x01\x12\x11\n\treference\x18\x1e \x01(\t\x12:\n\rcontact_above\x18( \x01(\x0b\x32#.kpex.tech.ProcessStackInfo.Contact\x1a\'\n\x0f\x46ieldOxideLayer\x12\x14\n\x0c\x64ielectric_k\x18\n \x01(\x01\x1a@\n\x15SimpleDielectricLayer\x12\x14\n\x0c\x64ielectric_k\x18\n \x01(\x01\x12\x11\n\treference\x18\x1e \x01(\t\x1a\x9f\x01\n\x18\x43onformalDielectricLayer\x12\x14\n\x0c\x64ielectric_k\x18\n \x01(\x01\x12\x1c\n\x14thickness_over_metal\x18\x14 \x01(\x01\x12 \n\x18thickness_where_no_metal\x18\x15 \x01(\x01\x12\x1a\n\x12thickness_sidewall\x18\x16 \x01(\x01\x12\x11\n\treference\x18\x1e \x01(\t\x1a~\n\x17SidewallDielectricLayer\x12\x14\n\x0c\x64ielectric_k\x18\n \x01(\x01\x12\x1a\n\x12height_above_metal\x18\x14 \x01(\x01\x12\x1e\n\x16width_outside_sidewall\x18\x15 \x01(\x01\x12\x11\n\treference\x18\x1e \x01(\t\x1a\x9d\x01\n\nMetalLayer\x12\x0e\n\x06height\x18\x01 \x01(\x01\x12\x11\n\tthickness\x18\x02 \x01(\x01\x12\x17\n\x0freference_below\x18\x1e \x01(\t\x12\x17\n\x0freference_above\x18\x1f \x01(\t\x12:\n\rcontact_above\x18( \x01(\x0b\x32#.kpex.tech.ProcessStackInfo.Contact\x1a\xc4\x05\n\tLayerInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x39\n\nlayer_type\x18\x02 \x01(\x0e\x32%.kpex.tech.ProcessStackInfo.LayerType\x12\x45\n\x0fsubstrate_layer\x18Z \x01(\x0b\x32*.kpex.tech.ProcessStackInfo.SubstrateLayerH\x00\x12=\n\x0bnwell_layer\x18\t \x01(\x0b\x32&.kpex.tech.ProcessStackInfo.NWellLayerH\x00\x12\x45\n\x0f\x64iffusion_layer\x18\n \x01(\x0b\x32*.kpex.tech.ProcessStackInfo.DiffusionLayerH\x00\x12H\n\x11\x66ield_oxide_layer\x18\x0b \x01(\x0b\x32+.kpex.tech.ProcessStackInfo.FieldOxideLayerH\x00\x12T\n\x17simple_dielectric_layer\x18\x0c \x01(\x0b\x32\x31.kpex.tech.ProcessStackInfo.SimpleDielectricLayerH\x00\x12Z\n\x1a\x63onformal_dielectric_layer\x18\r \x01(\x0b\x32\x34.kpex.tech.ProcessStackInfo.ConformalDielectricLayerH\x00\x12X\n\x19sidewall_dielectric_layer\x18\x0e \x01(\x0b\x32\x33.kpex.tech.ProcessStackInfo.SidewallDielectricLayerH\x00\x12=\n\x0bmetal_layer\x18\x0f \x01(\x0b\x32&.kpex.tech.ProcessStackInfo.MetalLayerH\x00\x42\x0c\n\nparameters\"\x8e\x02\n\tLayerType\x12\x1a\n\x16LAYER_TYPE_UNSPECIFIED\x10\x00\x12\x18\n\x14LAYER_TYPE_SUBSTRATE\x10\n\x12\x14\n\x10LAYER_TYPE_NWELL\x10\x14\x12\x18\n\x14LAYER_TYPE_DIFFUSION\x10\x1e\x12\x1a\n\x16LAYER_TYPE_FIELD_OXIDE\x10(\x12 \n\x1cLAYER_TYPE_SIMPLE_DIELECTRIC\x10\x32\x12#\n\x1fLAYER_TYPE_CONFORMAL_DIELECTRIC\x10<\x12\"\n\x1eLAYER_TYPE_SIDEWALL_DIELECTRIC\x10\x46\x12\x14\n\x10LAYER_TYPE_METAL\x10Pb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'process_stack_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_PROCESSSTACKINFO']._serialized_start=35
  _globals['_PROCESSSTACKINFO']._serialized_end=2008
  _globals['_PROCESSSTACKINFO_CONTACT']._serialized_start=110
  _globals['_PROCESSSTACKINFO_CONTACT']._serialized_end=173
  _globals['_PROCESSSTACKINFO_SUBSTRATELAYER']._serialized_start=175
  _globals['_PROCESSSTACKINFO_SUBSTRATELAYER']._serialized_end=245
  _globals['_PROCESSSTACKINFO_NWELLLAYER']._serialized_start=247
  _globals['_PROCESSSTACKINFO_NWELLLAYER']._serialized_end=354
  _globals['_PROCESSSTACKINFO_DIFFUSIONLAYER']._serialized_start=356
  _globals['_PROCESSSTACKINFO_DIFFUSIONLAYER']._serialized_end=467
  _globals['_PROCESSSTACKINFO_FIELDOXIDELAYER']._serialized_start=469
  _globals['_PROCESSSTACKINFO_FIELDOXIDELAYER']._serialized_end=508
  _globals['_PROCESSSTACKINFO_SIMPLEDIELECTRICLAYER']._serialized_start=510
  _globals['_PROCESSSTACKINFO_SIMPLEDIELECTRICLAYER']._serialized_end=574
  _globals['_PROCESSSTACKINFO_CONFORMALDIELECTRICLAYER']._serialized_start=577
  _globals['_PROCESSSTACKINFO_CONFORMALDIELECTRICLAYER']._serialized_end=736
  _globals['_PROCESSSTACKINFO_SIDEWALLDIELECTRICLAYER']._serialized_start=738
  _globals['_PROCESSSTACKINFO_SIDEWALLDIELECTRICLAYER']._serialized_end=864
  _globals['_PROCESSSTACKINFO_METALLAYER']._serialized_start=867
  _globals['_PROCESSSTACKINFO_METALLAYER']._serialized_end=1024
  _globals['_PROCESSSTACKINFO_LAYERINFO']._serialized_start=1027
  _globals['_PROCESSSTACKINFO_LAYERINFO']._serialized_end=1735
  _globals['_PROCESSSTACKINFO_LAYERTYPE']._serialized_start=1738
  _globals['_PROCESSSTACKINFO_LAYERTYPE']._serialized_end=2008
# @@protoc_insertion_point(module_scope)



#! /usr/bin/env python3
#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#

import argparse
from enum import Enum, StrEnum
from typing import *


def render_enum_help(topic: str,
                     enum_cls: Type[Enum],
                     print_default: bool = True,
                     lowercase_strenum: bool = False) -> str:
    def canonic_string(name: str, member: str) -> str:
        if issubclass(enum_cls, StrEnum):
            if name.lower() == 'default':
                return 'default'
            return member.lower() if lowercase_strenum else member
        return name.lower()
    if not hasattr(enum_cls, 'DEFAULT'):
        print_default = False
    case_list = [f"'{canonic_string(name, member)}'"
                 for name, member in enum_cls.__members__.items()
                 if name.lower() != 'default']
    enum_help = f"{topic} ∈ \u007b{', '.join(case_list)}\u007d"
    if print_default:
        default_case: enum_cls = getattr(enum_cls, 'DEFAULT')
        if issubclass(enum_cls, StrEnum):
            default_value: str = default_case.value
            if lowercase_strenum:
                default_value = default_value.lower()
        else:
            default_value = default_case.name.lower()
        enum_help += f".\nDefaults to '{default_value}'"
    return enum_help


def true_or_false(arg) -> bool:
    if isinstance(arg, bool):
        return arg

    match str(arg).lower():
        case 'yes' | 'true' | 't' | 'y' | 1:
            return True
        case 'no' | 'false' | 'f' | 'n' | 0:
            return False
        case _:
            raise argparse.ArgumentTypeError('Boolean value expected.')



#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
import re
import subprocess
import time
from typing import *

from ..log import (
    info,
    # warning,
    rule,
    subproc,
)
from ..common.capacitance_matrix import CapacitanceMatrix


def run_fastercap(exe_path: str,
                  lst_file_path: str,
                  log_path: str,
                  tolerance: float,
                  d_coeff: float,
                  mesh_refinement_value: float,
                  ooc_condition: Optional[int],
                  auto_preconditioner: bool,
                  galerkin_scheme: bool,
                  jacobi_preconditioner: bool):
    args = [
        exe_path,
        '-b',                          # console mode, without GUI
        '-i',                          # Dump detailed time and memory information
        '-v',                          # Verbose output
        f"-a{tolerance}",              # stop when relative error lower than threshold
        f"-d{d_coeff}",                # Direct potential interaction coefficient to mesh refinement ratio
        f"-m{mesh_refinement_value}",  # Mesh relative refinement value
    ]

    if ooc_condition is not None:
        args += [f"-f{ooc_condition}"]
    
    if auto_preconditioner:
        args += ['-ap']

    if galerkin_scheme:
        args += ['-g']

    if jacobi_preconditioner:
        args += ['-pj']

    args += [
        lst_file_path
    ]
    info(f"Calling FasterCap")
    subproc(f"{' '.join(args)}, output file: {log_path}")

    rule('FasterCap Output')
    start = time.time()

    proc = subprocess.Popen(args,
                            stdin=subprocess.DEVNULL,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            universal_newlines=True,
                            text=True)
    with open(log_path, 'w') as f:
        while True:
            line = proc.stdout.readline()
            if not line:
                break
            subproc(line[:-1])  # remove newline
            f.writelines([line])
    proc.wait()

    duration = time.time() - start

    rule()

    if proc.returncode == 0:
        info(f"FasterCap succeeded after {'%.4g' % duration}s")
    else:
        raise Exception(f"FasterCap failed with status code {proc.returncode} after {'%.4g' % duration}s, "
                        f"see log file: {log_path}")


def fastercap_parse_capacitance_matrix(log_path: str) -> CapacitanceMatrix:
    with open(log_path, 'r') as f:
        rlines = f.readlines()
        rlines.reverse()

        # multiple iterations possible, find the last matrix
        for idx, line in enumerate(rlines):
            if line.strip() == "Capacitance matrix is:":
                m = re.match(r'^Dimension (\d+) x (\d+)$', rlines[idx-1])
                if not m:
                    raise Exception(f"Could not parse capacitor matrix dimensions")
                dim = int(m.group(1))
                conductor_names: List[str] = []
                rows: List[List[float]] = []
                for i in reversed(range(idx-1-dim, idx-1)):
                    line = rlines[i].strip()
                    cells = [cell.strip() for cell in line.split(' ')]
                    cells = list(filter(lambda c: len(c) >= 1, cells))
                    conductor_names.append(cells[0])
                    row = [float(cell)/1e6 for cell in cells[1:]]
                    rows.append(row)
                cm = CapacitanceMatrix(conductor_names=conductor_names, rows=rows)
                return cm

        raise Exception(f"Could not extract capacitance matrix from FasterCap log file {log_path}")



#! /usr/bin/env python3
#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#

import argparse
from datetime import datetime
from enum import StrEnum
from functools import cached_property
import logging
import os
import os.path

import rich.console
import rich.markdown
import rich.text
from rich_argparse import RichHelpFormatter
import shlex
import shutil
import sys
from typing import *

import klayout.db as kdb
import klayout.rdb as rdb

from .fastercap.fastercap_input_builder import FasterCapInputBuilder
from .fastercap.fastercap_model_generator import FasterCapModelGenerator
from .fastercap.fastercap_runner import run_fastercap, fastercap_parse_capacitance_matrix
from .fastcap.fastcap_runner import run_fastcap, fastcap_parse_capacitance_matrix
from .klayout.lvs_runner import LVSRunner
from .klayout.lvsdb_extractor import KLayoutExtractionContext, KLayoutExtractedLayerInfo
from .klayout.netlist_expander import NetlistExpander
from .klayout.netlist_csv import NetlistCSVWriter
from .klayout.netlist_reducer import NetlistReducer
from .klayout.repair_rdb import repair_rdb
from .log import (
    LogLevel,
    set_log_level,
    register_additional_handler,
    deregister_additional_handler,
    # console,
    # debug,
    info,
    warning,
    subproc,
    error,
    rule
)
from .magic.magic_runner import MagicPEXMode, run_magic, prepare_magic_script
from .magic.magic_log_analyzer import MagicLogAnalyzer
from .pdk_config import PDKConfig
from .rcx25.extractor import RCExtractor, ExtractionResults
from .tech_info import TechInfo
from .util.multiple_choice import MultipleChoicePattern
from .util.argparse_helpers import render_enum_help, true_or_false
from .version import __version__


# ------------------------------------------------------------------------------------

PROGRAM_NAME = "kpex"


class ArgumentValidationError(Exception):
    pass


class InputMode(StrEnum):
    LVSDB = "lvsdb"
    GDS = "gds"


# TODO: this should be externally configurable
class PDK(StrEnum):
    IHP_SG13G2 = 'ihp_sg13g2'
    SKY130A = 'sky130A'

    @cached_property
    def config(self) -> PDKConfig:
        # NOTE: installation paths of resources in the distribution wheel differs from source repo
        base_dir = os.path.dirname(os.path.realpath(__file__))

        # NOTE: .git can be dir (standalone clone), or file (in case of submodule)
        if os.path.exists(os.path.join(base_dir, '..', '.git')): # in source repo
            base_dir = os.path.dirname(base_dir)
            tech_pb_json_dir = os.path.join(base_dir, 'klayout_pex_protobuf')
        else:  # site-packages/klayout_pex -> site-packages/klayout_pex_protobuf
            tech_pb_json_dir = os.path.join(os.path.dirname(base_dir), 'klayout_pex_protobuf')

        match self:
            case PDK.IHP_SG13G2:
                return PDKConfig(
                    name=self,
                    pex_lvs_script_path=os.path.join(base_dir, 'pdk', self, 'libs.tech', 'kpex', 'sg130g2.lvs'),
                    tech_pb_json_path=os.path.join(tech_pb_json_dir, f"{self}_tech.pb.json")
                )
            case PDK.SKY130A:
                return PDKConfig(
                    name=self,
                    pex_lvs_script_path=os.path.join(base_dir, 'pdk', self, 'libs.tech', 'kpex', 'sky130.lvs'),
                    tech_pb_json_path=os.path.join(tech_pb_json_dir, f"{self}_tech.pb.json")
                )


class KpexCLI:
    @staticmethod
    def parse_args(arg_list: List[str] = None) -> argparse.Namespace:
        # epilog = f"See '{PROGRAM_NAME} <subcommand> -h' for help on subcommand"
        epilog = """
| Variable | Example              | Description                             |
| -------- | -------------------- | --------------------------------------- |
| PDKPATH  | (e.g. $HOME/.volare) | Optional (required for default magicrc) |
| PDK      | (e.g. sky130A)       | Optional (required for default magicrc) |
"""
        epilog_md = rich.console.Group(
            rich.text.Text('Environmental variables:', style='argparse.groups'),
            rich.markdown.Markdown(epilog, style='argparse.text')
        )
        main_parser = argparse.ArgumentParser(description=f"{PROGRAM_NAME}: "
                                                          f"KLayout-integrated Parasitic Extraction Tool",
                                              epilog=epilog_md,
                                              add_help=False,
                                              formatter_class=RichHelpFormatter)

        group_special = main_parser.add_argument_group("Special options")
        group_special.add_argument("--help", "-h", action='help', help="show this help message and exit")
        group_special.add_argument("--version", "-v", action='version', version=f'{PROGRAM_NAME} {__version__}')
        group_special.add_argument("--log_level", dest='log_level', default='subprocess',
                                   help=render_enum_help(topic='log_level', enum_cls=LogLevel))
        group_special.add_argument("--threads", dest='num_threads', type=int,
                                   default=os.cpu_count() * 4,
                                   help="number of threads (e.g. for FasterCap) (default is %(default)s)")

        klayout_exe_default = 'klayout'
        if os.name == 'nt':
            klayout_exe_default = 'klayout_app'
        
        group_special.add_argument('--klayout', dest='klayout_exe_path', default=klayout_exe_default,
                                   help="Path to klayout executable (default is '%(default)s')")

        group_pex = main_parser.add_argument_group("Parasitic Extraction Setup")
        group_pex.add_argument("--pdk", dest="pdk", required=True,
                               type=PDK, choices=list(PDK),
                               help=render_enum_help(topic='pdk', enum_cls=PDK))

        group_pex.add_argument("--out_dir", "-o", dest="output_dir_base_path", default="output",
                               help="Output directory path (default is '%(default)s')")

        group_pex_input = main_parser.add_argument_group("Parasitic Extraction Input",
                                                         description="Either LVS is run, or an existing LVSDB is used")
        group_pex_input.add_argument("--gds", "-g", dest="gds_path", help="GDS path (for LVS)")
        group_pex_input.add_argument("--schematic", "-s", dest="schematic_path",
                                     help="Schematic SPICE netlist path (for LVS). "
                                          "If none given, a dummy schematic will be created")
        group_pex_input.add_argument("--lvsdb", "-l", dest="lvsdb_path", help="KLayout LVSDB path (bypass LVS)")
        group_pex_input.add_argument("--cell", "-c", dest="cell_name", default=None,
                                     help="Cell (default is the top cell)")

        group_pex_input.add_argument("--cache-lvs", dest="cache_lvs",
                                     type=true_or_false, default=True,
                                     help="Used cached LVSDB (for given input GDS) (default is %(default)s)")
        group_pex_input.add_argument("--cache-dir", dest="cache_dir_path", default=None,
                                     help="Path for cached LVSDB (default is .kpex_cache within --out_dir)")
        group_pex_input.add_argument("--lvs-verbose", dest="klayout_lvs_verbose",
                                     type=true_or_false, default=False,
                                     help="Verbose KLayout LVS output (default is %(default)s)")

        group_pex_options = main_parser.add_argument_group("Parasitic Extraction Options")
        group_pex_options.add_argument("--blackbox", dest="blackbox_devices",
                                      type=true_or_false, default=False,  # TODO: in the future this should be True by default
                                      help="Blackbox devices like MIM/MOM caps, as they are handled by SPICE models "
                                           "(default is %(default)s for testing now)")
        group_pex_options.add_argument("--fastercap", dest="run_fastercap",
                                      action='store_true', default=False,
                                      help="Run FasterCap engine (default is %(default)s)")
        group_pex_options.add_argument("--fastcap", dest="run_fastcap",
                                      action='store_true', default=False,
                                      help="Run FastCap2 engine (default is %(default)s)")
        group_pex_options.add_argument("--magic", dest="run_magic",
                                      action='store_true', default=False,
                                      help="Run MAGIC engine (default is %(default)s)")
        group_pex_options.add_argument("--2.5D", dest="run_2_5D",
                                      action='store_true', default=False,
                                      help="Run 2.5D analytical engine (default is %(default)s)")

        group_fastercap = main_parser.add_argument_group("FasterCap options")
        group_fastercap.add_argument("--k_void", "-k", dest="k_void",
                                     type=float, default=3.9,
                                     help="Dielectric constant of void (default is %(default)s)")
        group_fastercap.add_argument("--delaunay_amax", "-a", dest="delaunay_amax",
                                     type=float, default=50,
                                     help="Delaunay triangulation maximum area (default is %(default)s)")
        group_fastercap.add_argument("--delaunay_b", "-b", dest="delaunay_b",
                                     type=float, default=0.5,
                                     help="Delaunay triangulation b (default is %(default)s)")
        group_fastercap.add_argument("--geo_check", dest="geometry_check",
                                     type=true_or_false, default=False,
                                     help=f"Validate geometries before passing to FasterCap "
                                          f"(default is False)")
        group_fastercap.add_argument("--diel", dest="dielectric_filter",
                                     type=str, default="all",
                                     help=f"Comma separated list of dielectric filter patterns. "
                                          f"Allowed patterns are: (none, all, -dielname1, +dielname2) "
                                          f"(default is %(default)s)")

        group_fastercap.add_argument("--tolerance", dest="fastercap_tolerance",
                                     type=float, default=0.05,
                                     help="FasterCap -aX error tolerance (default is %(default)s)")
        group_fastercap.add_argument("--d_coeff", dest="fastercap_d_coeff",
                                     type=float, default=0.5,
                                     help=f"FasterCap -d direct potential interaction coefficient to mesh refinement "
                                          f"(default is %(default)s)")
        group_fastercap.add_argument("--mesh", dest="fastercap_mesh_refinement_value",
                                     type=float, default=0.5,
                                     help="FasterCap -m Mesh relative refinement value (default is %(default)s)")
        group_fastercap.add_argument("--ooc", dest="fastercap_ooc_condition",
                                     type=float, default=2,
                                     help="FasterCap -f out-of-core free memory to link memory condition "
                                          "(0 = don't go OOC, default is %(default)s)")
        group_fastercap.add_argument("--auto_precond", dest="fastercap_auto_preconditioner",
                                     type=true_or_false, default=True,
                                     help=f"FasterCap -ap Automatic preconditioner usage (default is %(default)s)")
        group_fastercap.add_argument("--galerkin", dest="fastercap_galerkin_scheme",
                                     action='store_true', default=False,
                                     help=f"FasterCap -g Use Galerkin scheme (default is %(default)s)")
        group_fastercap.add_argument("--jacobi", dest="fastercap_jacobi_preconditioner",
                                     action='store_true', default=False,
                                     help="FasterCap -pj Use Jacobi preconditioner (default is %(default)s)")

        PDKPATH = os.environ.get('PDKPATH', None)
        default_magicrc_path = \
            None if PDKPATH is None \
            else os.path.abspath(f"{PDKPATH}/libs.tech/magic/{os.environ['PDK']}.magicrc")
        group_magic = main_parser.add_argument_group("MAGIC options")
        group_magic.add_argument('--magicrc', dest='magicrc_path', default=default_magicrc_path,
                                  help=f"Path to magicrc configuration file (default is '%(default)s')")
        group_magic.add_argument("--magic_mode", dest='magic_pex_mode',
                                 default=MagicPEXMode.DEFAULT, type=MagicPEXMode, choices=list(MagicPEXMode),
                                 help=render_enum_help(topic='magic_mode', enum_cls=MagicPEXMode))
        group_magic.add_argument("--magic_cthresh", dest="magic_cthresh",
                                 type=float, default=0.01,
                                 help="Threshold (in fF) for ignored parasitic capacitances (default is %(default)s). "
                                      "(MAGIC command: ext2spice cthresh <value>)")
        group_magic.add_argument("--magic_rthresh", dest="magic_rthresh",
                                 type=int, default=100,
                                 help="Threshold (in Ω) for ignored parasitic resistances (default is %(default)s). "
                                      "(MAGIC command: ext2spice rthresh <value>)")
        group_magic.add_argument("--magic_tolerance", dest="magic_tolerance",
                                 type=float, default=1,
                                 help="Set ratio between resistor and device tolerance (default is %(default)s). "
                                      "(MAGIC command: extresist tolerance <value>)")
        group_magic.add_argument("--magic_halo", dest="magic_halo",
                                 type=float, default=None,
                                 help="Custom sidewall halo distance (in µm) "
                                      "(MAGIC command: extract halo <value>) (default is no custom halo)")
        group_magic.add_argument('--magic_exe', dest='magic_exe_path', default='magic',
                                  help="Path to magic executable (default is '%(default)s')")

        if arg_list is None:
            arg_list = sys.argv[1:]
        args = main_parser.parse_args(arg_list)
        return args

    @staticmethod
    def validate_args(args: argparse.Namespace):
        found_errors = False

        pdk_config: PDKConfig = args.pdk.config
        args.tech_pbjson_path = pdk_config.tech_pb_json_path
        args.lvs_script_path = pdk_config.pex_lvs_script_path

        if not os.path.isfile(args.klayout_exe_path):
            path = shutil.which(args.klayout_exe_path)
            if not path:
                error(f"Can't locate KLayout executable at {args.klayout_exe_path}")
                found_errors = True

        if not os.path.isfile(args.tech_pbjson_path):
            error(f"Can't read technology file at path {args.tech_pbjson_path}")
            found_errors = True

        if not os.path.isfile(args.lvs_script_path):
            error(f"Can't locate LVS script path at {args.lvs_script_path}")
            found_errors = True

        rule('Input Layout')

        # input mode: LVS or existing LVSDB?
        if args.gds_path:
            info(f"GDS input file passed, running in LVS mode")
            args.input_mode = InputMode.GDS
            if not os.path.isfile(args.gds_path):
                error(f"Can't read GDS file (LVS input) at path {args.gds_path}")
                found_errors = True
            else:
                args.layout = kdb.Layout()
                args.layout.read(args.gds_path)

                top_cells = args.layout.top_cells()

                if args.cell_name:  # explicit user-specified cell name
                    args.effective_cell_name = args.cell_name

                    found_cell: Optional[kdb.Cell] = None
                    for cell in args.layout.cells('*'):
                        if cell.name == args.effective_cell_name:
                            found_cell = cell
                            break
                    if not found_cell:
                        error(f"Could not find cell {args.cell_name} in GDS {args.gds_path}")
                        found_errors = True

                    is_only_top_cell = len(top_cells) == 1 and top_cells[0].name == args.cell_name
                    if is_only_top_cell:
                        info(f"Found cell {args.cell_name} in GDS {args.gds_path} (only top cell)")
                    else:  # there are other cells => extract the top cell to a tmp layout
                        args.effective_gds_path = os.path.join(args.output_dir_path, f"{args.cell_name}_exported.gds.gz")
                        info(f"Found cell {args.cell_name} in GDS {args.gds_path}, "
                             f"but it is not the only top cell, "
                             f"so layout is exported to: {args.effective_gds_path}")

                        found_cell.write(args.effective_gds_path)
                else:  # find top cell
                    if len(top_cells) == 1:
                        args.effective_cell_name = top_cells[0].name
                        info(f"No explicit top cell specified, using top cell '{args.effective_cell_name}'")
                    else:
                        args.effective_cell_name = 'TOP'
                        error(f"Could not determine the default top cell in GDS {args.gds_path}, "
                              f"there are multiple: {', '.join([c.name for c in top_cells])}. "
                              f"Use --cell to specify the cell")
                        found_errors = True

                args.effective_gds_path = args.gds_path
        else:
            info(f"LVSDB input file passed, bypassing LVS")
            args.input_mode = InputMode.LVSDB
            if not hasattr(args, 'lvsdb_path'):
                error(f"LVSDB input path not specified (argument --lvsdb)")
                found_errors = True
            elif not os.path.isfile(args.lvsdb_path):
                error(f"Can't read KLayout LVSDB file at path {args.lvsdb_path}")
                found_errors = True
            else:
                lvsdb = kdb.LayoutVsSchematic()
                lvsdb.read(args.lvsdb_path)
                top_cell: kdb.Cell = lvsdb.internal_top_cell()
                args.effective_cell_name = top_cell.name

        def input_file_stem(path: str):
            # could be *.gds, or *.gds.gz, so remove all extensions
            return os.path.basename(path).split(sep='.')[0]

        if hasattr(args, 'effective_cell_name'):
            run_dir_id: str
            match args.input_mode:
                case InputMode.GDS:
                    run_dir_id = f"{input_file_stem(args.gds_path)}__{args.effective_cell_name}"
                case InputMode.LVSDB:
                    run_dir_id = f"{input_file_stem(args.lvsdb_path)}__{args.effective_cell_name}"
                case _:
                    raise NotImplementedError(f"Unknown input mode {args.input_mode}")

            args.output_dir_path = os.path.join(args.output_dir_base_path, run_dir_id)
            os.makedirs(args.output_dir_path, exist_ok=True)
            if args.input_mode == InputMode.GDS:
                if args.schematic_path:
                    args.effective_schematic_path = args.schematic_path
                    if not os.path.isfile(args.schematic_path):
                        error(f"Can't read schematic (LVS input) at path {args.schematic_path}")
                        found_errors = True
                else:
                    info(f"LVS input schematic not specified (argument --schematic), using dummy schematic")
                    args.effective_schematic_path = os.path.join(args.output_dir_path,
                                                                 f"{args.effective_cell_name}_dummy_schematic.spice")
                    with open(args.effective_schematic_path, 'w') as f:
                        f.writelines([
                            f".subckt {args.effective_cell_name} VDD VSS",
                            '.ends',
                            '.end'
                        ])

        try:
            args.log_level = LogLevel[args.log_level.upper()]
        except KeyError:
            error(f"Requested log level {args.log_level.lower()} does not exist, "
                  f"{render_enum_help(topic='log_level', enum_cls=LogLevel, print_default=False)}")
            found_errors = True

        try:
            pattern_string: str = args.dielectric_filter
            args.dielectric_filter = MultipleChoicePattern(pattern=pattern_string)
        except ValueError as e:
            error("Failed to parse --diel arg", e)
            found_errors = True

        # at least one engine must be activated

        if not (args.run_magic or args.run_fastcap or args.run_fastercap or args.run_2_5D):
            error("No PEX engines activated")
            engine_help = """
| Argument     | Description               |
| ------------ | ------------------------- |
| --fastercap  | Run kpex/FasterCap engine |
| --2.5D       | Run kpex/2.5D engine      |
| --magic      | Run MAGIC engine          |
"""
            subproc(f"\n\nPlease activate one or more engines using the arguments:")
            rich.print(rich.markdown.Markdown(engine_help, style='argparse.text'))
            found_errors = True

        if args.cache_dir_path is None:
            args.cache_dir_path = os.path.join(args.output_dir_base_path, '.kpex_cache')

        if found_errors:
            raise ArgumentValidationError("Argument validation failed")

    def build_fastercap_input(self,
                              args: argparse.Namespace,
                              pex_context: KLayoutExtractionContext,
                              tech_info: TechInfo) -> str:
        rule('Process stackup')
        fastercap_input_builder = FasterCapInputBuilder(pex_context=pex_context,
                                                        tech_info=tech_info,
                                                        k_void=args.k_void,
                                                        delaunay_amax=args.delaunay_amax,
                                                        delaunay_b=args.delaunay_b)
        gen: FasterCapModelGenerator = fastercap_input_builder.build()

        rule('FasterCap Input File Generation')
        faster_cap_input_dir_path = os.path.join(args.output_dir_path, 'FasterCap_Input_Files')
        os.makedirs(faster_cap_input_dir_path, exist_ok=True)

        lst_file = gen.write_fastcap(output_dir_path=faster_cap_input_dir_path, prefix='FasterCap_Input_')

        rule('STL File Generation')
        geometry_dir_path = os.path.join(args.output_dir_path, 'Geometries')
        os.makedirs(geometry_dir_path, exist_ok=True)
        gen.dump_stl(output_dir_path=geometry_dir_path, prefix='')

        if args.geometry_check:
            rule('Geometry Validation')
            gen.check()

        return lst_file


    def run_fastercap_extraction(self,
                                 args: argparse.Namespace,
                                 pex_context: KLayoutExtractionContext,
                                 lst_file: str):
        rule('FasterCap Execution')
        info(f"Configure number of OpenMP threads (environmental variable OMP_NUM_THREADS) as {args.num_threads}")
        os.environ['OMP_NUM_THREADS'] = f"{args.num_threads}"

        exe_path = "FasterCap"
        log_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_FasterCap_Output.txt")
        raw_csv_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_FasterCap_Result_Matrix_Raw.csv")
        avg_csv_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_FasterCap_Result_Matrix_Avg.csv")
        expanded_netlist_path = os.path.join(args.output_dir_path,
                                             f"{args.effective_cell_name}_FasterCap_Expanded_Netlist.cir")
        expanded_netlist_csv_path = os.path.join(args.output_dir_path,
                                                 f"{args.effective_cell_name}_FasterCap_Expanded_Netlist.csv")
        reduced_netlist_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_FasterCap_Reduced_Netlist.cir")

        run_fastercap(exe_path=exe_path,
                      lst_file_path=lst_file,
                      log_path=log_path,
                      tolerance=args.fastercap_tolerance,
                      d_coeff=args.fastercap_d_coeff,
                      mesh_refinement_value=args.fastercap_mesh_refinement_value,
                      ooc_condition=args.fastercap_ooc_condition,
                      auto_preconditioner=args.fastercap_auto_preconditioner,
                      galerkin_scheme=args.fastercap_galerkin_scheme,
                      jacobi_preconditioner=args.fastercap_jacobi_preconditioner)

        cap_matrix = fastercap_parse_capacitance_matrix(log_path)
        cap_matrix.write_csv(raw_csv_path)

        cap_matrix = cap_matrix.averaged_off_diagonals()
        cap_matrix.write_csv(avg_csv_path)

        netlist_expander = NetlistExpander()
        expanded_netlist = netlist_expander.expand(
            extracted_netlist=pex_context.lvsdb.netlist(),
            top_cell_name=pex_context.top_cell.name,
            cap_matrix=cap_matrix,
            blackbox_devices=args.blackbox_devices
        )

        # create a nice CSV for reports, useful for spreadsheets
        netlist_csv_writer = NetlistCSVWriter()
        netlist_csv_writer.write_csv(netlist=expanded_netlist,
                                     top_cell_name=pex_context.top_cell.name,
                                     output_path=expanded_netlist_csv_path)

        rule("Extended netlist (CSV format):")
        with open(expanded_netlist_csv_path, 'r') as f:
            for line in f.readlines():
                subproc(line[:-1])  # abusing subproc, simply want verbatim
        rule()

        info(f"Wrote expanded netlist CSV to: {expanded_netlist_csv_path}")

        spice_writer = kdb.NetlistSpiceWriter()
        spice_writer.use_net_names = True
        spice_writer.with_comments = False
        expanded_netlist.write(expanded_netlist_path, spice_writer)
        info(f"Wrote expanded netlist to: {expanded_netlist_path}")

        netlist_reducer = NetlistReducer()
        reduced_netlist = netlist_reducer.reduce(netlist=expanded_netlist,
                                                 top_cell_name=pex_context.top_cell.name)
        reduced_netlist.write(reduced_netlist_path, spice_writer)
        info(f"Wrote reduced netlist to: {reduced_netlist_path}")

        self._fastercap_extracted_csv_path = expanded_netlist_csv_path

    def run_magic_extraction(self,
                             args: argparse.Namespace):
        if args.input_mode != InputMode.GDS:
            error(f"MAGIC engine only works with GDS input mode"
                  f" (currently {args.input_mode})")
            return

        magic_run_dir = os.path.join(args.output_dir_path, f"magic_{args.magic_pex_mode}")
        magic_log_path = os.path.join(magic_run_dir,
                                      f"{args.effective_cell_name}_MAGIC_{args.magic_pex_mode}_Output.txt")
        magic_script_path = os.path.join(magic_run_dir,
                                         f"{args.effective_cell_name}_MAGIC_{args.magic_pex_mode}_Script.tcl")

        output_netlist_path = os.path.join(magic_run_dir, f"{args.effective_cell_name}.pex.spice")
        report_db_path = os.path.join(magic_run_dir, f"{args.effective_cell_name}_MAGIC_report.rdb.gz")

        os.makedirs(magic_run_dir, exist_ok=True)

        prepare_magic_script(gds_path=args.effective_gds_path,
                             cell_name=args.effective_cell_name,
                             run_dir_path=magic_run_dir,
                             script_path=magic_script_path,
                             output_netlist_path=output_netlist_path,
                             pex_mode=args.magic_pex_mode,
                             c_threshold=args.magic_cthresh,
                             r_threshold=args.magic_rthresh,
                             tolerance=args.magic_tolerance,
                             halo=args.magic_halo)

        run_magic(exe_path=args.magic_exe_path,
                  magicrc_path=args.magicrc_path,
                  script_path=magic_script_path,
                  log_path=magic_log_path)

        layout = kdb.Layout()
        layout.read(args.effective_gds_path)

        report = rdb.ReportDatabase('')
        magic_log_analyzer = MagicLogAnalyzer(magic_log_dir_path=magic_run_dir,
                                              report=report,
                                              dbu=layout.dbu)
        magic_log_analyzer.analyze()
        report.save(report_db_path)

        rule("Paths")
        subproc(f"Report DB saved at: {report_db_path}")
        subproc(f"SPICE netlist saved at: {output_netlist_path}")

        rule("MAGIC PEX SPICE netlist")
        with open(output_netlist_path, 'r') as f:
            subproc(f.read())
        rule()

    def run_fastcap_extraction(self,
                               args: argparse.Namespace,
                               pex_context: KLayoutExtractionContext,
                               lst_file: str):
        rule('FastCap2 Execution')
        exe_path = "fastcap"
        log_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_FastCap2_Output.txt")
        raw_csv_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_FastCap2_Result_Matrix_Raw.csv")
        avg_csv_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_FastCap2_Result_Matrix_Avg.csv")
        expanded_netlist_path = os.path.join(args.output_dir_path,
                                             f"{args.effective_cell_name}_FastCap2_Expanded_Netlist.cir")
        reduced_netlist_path = os.path.join(args.output_dir_path,
                                            f"{args.effective_cell_name}_FastCap2_Reduced_Netlist.cir")

        run_fastcap(exe_path=exe_path,
                    lst_file_path=lst_file,
                    log_path=log_path)

        cap_matrix = fastcap_parse_capacitance_matrix(log_path)
        cap_matrix.write_csv(raw_csv_path)

        cap_matrix = cap_matrix.averaged_off_diagonals()
        cap_matrix.write_csv(avg_csv_path)

        netlist_expander = NetlistExpander()
        expanded_netlist = netlist_expander.expand(
            extracted_netlist=pex_context.lvsdb.netlist(),
            top_cell_name=pex_context.top_cell.name,
            cap_matrix=cap_matrix,
            blackbox_devices=args.blackbox_devices
        )

        spice_writer = kdb.NetlistSpiceWriter()
        spice_writer.use_net_names = True
        spice_writer.with_comments = False
        expanded_netlist.write(expanded_netlist_path, spice_writer)
        info(f"Wrote expanded netlist to: {expanded_netlist_path}")

        netlist_reducer = NetlistReducer()
        reduced_netlist = netlist_reducer.reduce(netlist=expanded_netlist,
                                                 top_cell_name=pex_context.top_cell.name)
        reduced_netlist.write(reduced_netlist_path, spice_writer)
        info(f"Wrote reduced netlist to: {reduced_netlist_path}")

    def run_kpex_2_5d_engine(self,
                             args: argparse.Namespace,
                             pex_context: KLayoutExtractionContext,
                             tech_info: TechInfo,
                             report_path: str,
                             netlist_csv_path: str):
        extractor = RCExtractor(pex_context=pex_context,
                                tech_info=tech_info,
                                report_path=report_path)
        extraction_results = extractor.extract()

        with open(netlist_csv_path, 'w') as f:
            f.write('Device;Net1;Net2;Capacitance [fF]\n')
            # f.write('Device;Net1;Net2;Capacitance [F];Capacitance [fF]\n')
            summary = extraction_results.summarize()
            for idx, (key, cap_value) in enumerate(summary.capacitances.items()):
                # f.write(f"C{idx + 1};{key.net1};{key.net2};{cap_value / 1e15};{round(cap_value, 3)}\n")
                f.write(f"C{idx + 1};{key.net1};{key.net2};{round(cap_value, 3)}\n")

        rule("kpex/2.5D extracted netlist (CSV format):")
        with open(netlist_csv_path, 'r') as f:
            for line in f.readlines():
                subproc(line[:-1])  # abusing subproc, simply want verbatim

        rule("Extracted netlist CSV")
        subproc(f"{netlist_csv_path}")


        # NOTE: there was a KLayout bug that some of the categories were lost,
        #       so that the marker browser could not load the report file
        try:
            report = rdb.ReportDatabase('')
            report.load(report_path)  # try loading rdb
        except Exception as e:
            rule("Repair broken marker DB")
            warning(f"Detected KLayout bug: RDB can't be loaded due to exception {e}")
            repair_rdb(report_path)

        return extraction_results

    def setup_logging(self, args: argparse.Namespace):
        def register_log_file_handler(log_path: str,
                                      formatter: Optional[logging.Formatter]) -> logging.Handler:
            handler = logging.FileHandler(log_path)
            handler.setLevel(LogLevel.SUBPROCESS)
            if formatter:
                handler.setFormatter(formatter)
            register_additional_handler(handler)
            return handler

        def reregister_log_file_handler(handler: logging.Handler,
                                        log_path: str,
                                        formatter: Optional[logging.Formatter]):
            deregister_additional_handler(handler)
            handler.flush()
            handler.close()
            os.makedirs(args.output_dir_path, exist_ok=True)
            new_path = os.path.join(args.output_dir_path, os.path.basename(log_path))
            if os.path.exists(new_path):
                ctime = os.path.getctime(new_path)
                dt = datetime.fromtimestamp(ctime)
                timestamp = dt.strftime('%Y-%m-%d_%H-%M-%S')
                backup_path = f"{new_path[:-4]}_{timestamp}.bak.log"
                shutil.move(new_path, backup_path)
            log_path = shutil.move(log_path, new_path)
            register_log_file_handler(log_path, formatter)

        # setup preliminary logger
        cli_log_path_plain = os.path.join(args.output_dir_base_path, f"kpex_plain.log")
        cli_log_path_formatted = os.path.join(args.output_dir_base_path, f"kpex.log")
        formatter = logging.Formatter('[%(asctime)s] [%(levelname)s]    %(message)s')
        file_handler_plain = register_log_file_handler(cli_log_path_plain, None)
        file_handler_formatted = register_log_file_handler(cli_log_path_formatted, formatter)
        try:
            self.validate_args(args)
        except ArgumentValidationError:
            if hasattr(args, 'output_dir_path'):
                reregister_log_file_handler(file_handler_plain, cli_log_path_plain, None)
                reregister_log_file_handler(file_handler_formatted, cli_log_path_formatted, formatter)
            sys.exit(1)
        reregister_log_file_handler(file_handler_plain, cli_log_path_plain, None)
        reregister_log_file_handler(file_handler_formatted, cli_log_path_formatted, formatter)

        set_log_level(args.log_level)

    @staticmethod
    def modification_date(filename: str) -> datetime:
        t = os.path.getmtime(filename)
        return datetime.fromtimestamp(t)

    def create_lvsdb(self, args: argparse.Namespace) -> kdb.LayoutVsSchematic:
        lvsdb = kdb.LayoutVsSchematic()

        match args.input_mode:
            case InputMode.LVSDB:
                lvsdb.read(args.lvsdb_path)
            case InputMode.GDS:
                lvs_log_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_lvs.log")
                lvsdb_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}.lvsdb.gz")
                lvsdb_cache_path = os.path.join(args.cache_dir_path, args.pdk,
                                                os.path.splitroot(os.path.abspath(args.gds_path))[-1],
                                                f"{args.effective_cell_name}.lvsdb.gz")

                lvs_needed = True

                if args.cache_lvs:
                    if not os.path.exists(lvsdb_cache_path):
                        info(f"Cache miss: extracted LVSDB does not exist")
                        subproc(lvsdb_cache_path)
                    elif self.modification_date(lvsdb_cache_path) <= self.modification_date(args.gds_path):
                        info(f"Cache miss: extracted LVSDB is older than the input GDS")
                        subproc(lvsdb_cache_path)
                    else:
                        warning(f"Cache hit: Reusing cached LVSDB")
                        subproc(lvsdb_cache_path)
                        lvs_needed = False

                if lvs_needed:
                    lvs_runner = LVSRunner()
                    lvs_runner.run_klayout_lvs(exe_path=args.klayout_exe_path,
                                               lvs_script=args.lvs_script_path,
                                               gds_path=args.effective_gds_path,
                                               schematic_path=args.effective_schematic_path,
                                               log_path=lvs_log_path,
                                               lvsdb_path=lvsdb_path,
                                               verbose=args.klayout_lvs_verbose)
                    if args.cache_lvs:
                        cache_dir_path = os.path.dirname(lvsdb_cache_path)
                        if not os.path.exists(cache_dir_path):
                            os.makedirs(cache_dir_path, exist_ok=True)
                        shutil.copy(lvsdb_path, lvsdb_cache_path)

                lvsdb.read(lvsdb_path)
        return lvsdb

    def main(self, argv: List[str]):
        if '-v' not in argv and \
           '--version' not in argv and \
           '-h' not in argv and \
           '--help' not in argv:
            rule('Command line arguments')
            subproc(' '.join(map(shlex.quote, sys.argv)))

        args = self.parse_args(argv[1:])

        os.makedirs(args.output_dir_base_path, exist_ok=True)
        self.setup_logging(args)

        tech_info = TechInfo.from_json(args.tech_pbjson_path,
                                       dielectric_filter=args.dielectric_filter)

        if args.run_magic:
            rule('MAGIC')
            self.run_magic_extraction(args)

        # no need to run LVS etc if only running magic engine
        if not (args.run_fastcap or args.run_fastercap or args.run_2_5D):
            return

        rule('Prepare LVSDB')
        lvsdb = self.create_lvsdb(args)

        pex_context = KLayoutExtractionContext.prepare_extraction(top_cell=args.effective_cell_name,
                                                                  lvsdb=lvsdb,
                                                                  tech=tech_info,
                                                                  blackbox_devices=args.blackbox_devices)
        rule('Non-empty layers in LVS database')
        for gds_pair, layer_info in pex_context.extracted_layers.items():
            names = [l.lvs_layer_name for l in layer_info.source_layers]
            info(f"{gds_pair} -> ({' '.join(names)})")

        gds_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_l2n_extracted.gds.gz")
        pex_context.target_layout.write(gds_path)

        gds_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_l2n_internal.gds.gz")
        pex_context.lvsdb.internal_layout().write(gds_path)

        def dump_layers(cell: str,
                        layers: List[KLayoutExtractedLayerInfo],
                        layout_dump_path: str):
            layout = kdb.Layout()
            layout.dbu = lvsdb.internal_layout().dbu

            top_cell = layout.create_cell(cell)
            for ulyr in layers:
                li = kdb.LayerInfo(*ulyr.gds_pair)
                li.name = ulyr.lvs_layer_name
                layer = layout.insert_layer(li)
                layout.insert(top_cell.cell_index(), layer, ulyr.region.dup())

            layout.write(layout_dump_path)

        if len(pex_context.unnamed_layers) >= 1:
            layout_dump_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_unnamed_LVS_layers.gds.gz")
            dump_layers(cell=args.effective_cell_name,
                        layers=pex_context.unnamed_layers,
                        layout_dump_path=layout_dump_path)

        if len(pex_context.extracted_layers) >= 1:
            layout_dump_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_nonempty_LVS_layers.gds.gz")
            nonempty_layers = [l \
                               for layers in pex_context.extracted_layers.values() \
                               for l in layers.source_layers]
            dump_layers(cell=args.effective_cell_name,
                        layers=nonempty_layers,
                        layout_dump_path=layout_dump_path)
        else:
            error("No extracted layers found")
            sys.exit(1)

        if args.run_fastcap or args.run_fastercap:
            lst_file = self.build_fastercap_input(args=args,
                                                  pex_context=pex_context,
                                                  tech_info=tech_info)
            if args.run_fastercap:
                self.run_fastercap_extraction(args=args,
                                              pex_context=pex_context,
                                              lst_file=lst_file)
            if args.run_fastcap:
                self.run_fastcap_extraction(args=args,
                                            pex_context=pex_context,
                                            lst_file=lst_file)

        if args.run_2_5D:
            rule("kpex/2.5D PEX Engine")
            report_path = os.path.join(args.output_dir_path, f"{args.effective_cell_name}_k25d_pex_report.rdb.gz")
            netlist_csv_path = os.path.abspath(os.path.join(args.output_dir_path, f"{args.effective_cell_name}_k25d_pex_netlist.csv"))

            self._rcx25_extraction_results = self.run_kpex_2_5d_engine(  # NOTE: store for test case
                args=args,
                pex_context=pex_context,
                tech_info=tech_info,
                report_path=report_path,
                netlist_csv_path=netlist_csv_path
            )

            self._rcx25_extracted_csv_path = netlist_csv_path

    @property
    def rcx25_extraction_results(self) -> ExtractionResults:
        if not hasattr(self, '_rcx25_extraction_results'):
            raise Exception('rcx25_extraction_results is not initialized, was run_kpex_2_5d_engine called?')
        return self._rcx25_extraction_results

    @property
    def rcx25_extracted_csv_path(self) -> str:
        if not hasattr(self, '_rcx25_extracted_csv_path'):
            raise Exception('rcx25_extracted_csv_path is not initialized, was run_kpex_2_5d_engine called?')
        return self._rcx25_extracted_csv_path

    @property
    def fastercap_extracted_csv_path(self) -> str:
        if not hasattr(self, '_fastercap_extracted_csv_path'):
            raise Exception('fastercap_extracted_csv_path is not initialized, was run_fastercap_extraction called?')
        return self._fastercap_extracted_csv_path


if __name__ == "__main__":
    cli = KpexCLI()
    cli.main(sys.argv)



#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
from __future__ import annotations
from typing import *


class MultipleChoicePattern:
    def __init__(self, pattern: str):
        """
        Multiple Choice pattern, allows blacklisting and whitelisting.
        For example, given a list of dielectric, let the user decide which of them to include or exclude.
        Allowed patterns:
            - all (default): complete list of choices included
            - none: no choices included at all
            - +dielname: include choice named 'dielname'
            - -dielname: exclude choice named 'dielname'
        Examples:
            - all,-nild5,-nild6
               - include all dielectrics except nild5 and nild6
            - none,+nild5,+capild
                - include only dielectrics named nild5 and capild
        """
        self.pattern = pattern

        components = pattern.split(sep=',')
        components = [c.lower().strip() for c in components]
        self.has_all = 'all' in components
        self.has_none = 'none' in components
        self.included = [c[1:] for c in components if c.startswith('+')]
        self.excluded = [c[1:] for c in components if c.startswith('-')]
        if self.has_none and self.has_all:
            raise ValueError("Multiple choice pattern can't have both subpatterns all and none")
        if self.has_none and len(self.excluded) >= 1:
            raise ValueError("Multiple choice pattern based on none can only have inclusive (+) subpatterns")
        if self.has_all and len(self.included) >= 1:
            raise ValueError("Multiple choice pattern based on all can only have exclusive (-) subpatterns")

    def filter(self, choices: List[str]) -> List[str]:
        if self.has_all:
            return [c for c in choices if c not in self.excluded]
        return [c for c in choices if c in self.included]

    def is_included(self, choice: str) -> bool:
        if self.has_none:
            return choice in self.included
        if self.has_all:
            return choice not in self.excluded
        return False


#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
from __future__ import annotations

import klayout.db as kdb

from klayout_pex.log import (
    info,
)


class NetlistCSVWriter:
    @staticmethod
    def write_csv(netlist: kdb.Netlist,
                  top_cell_name: str,
                  output_path: str):
        with open(output_path, 'w') as f:
            f.write('Device;Net1;Net2;Capacitance [fF]\n')

            top_circuit: kdb.Circuit = netlist.circuit_by_name(top_cell_name)

            # NOTE: only caps for now
            for d in top_circuit.each_device():
                # https://www.klayout.de/doc-qt5/code/class_Device.html
                dc = d.device_class()
                if isinstance(dc, kdb.DeviceClassCapacitor):
                    dn = d.expanded_name() or d.name
                    if dc.name != 'PEX_CAP':
                        info(f"Ignoring device {dn}")
                        continue
                    param_defs = dc.parameter_definitions()
                    params = {p.name: d.parameter(p.id()) for p in param_defs}
                    d: kdb.Device
                    net1 = d.net_for_terminal('A')
                    net2 = d.net_for_terminal('B')
                    cap = params['C']
                    cap_femto = round(cap * 1e15, 2)
                    f.write(f"{dn};{net1.name};{net2.name};{cap_femto}\n")



#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
from __future__ import annotations

import re
from typing import *

import klayout.db as kdb

from ..log import (
    info,
    warning,
)
from ..common.capacitance_matrix import CapacitanceMatrix


class NetlistExpander:
    @staticmethod
    def expand(extracted_netlist: kdb.Netlist,
               top_cell_name: str,
               cap_matrix: CapacitanceMatrix,
               blackbox_devices: bool) -> kdb.Netlist:
        expanded_netlist: kdb.Netlist = extracted_netlist.dup()
        top_circuit: kdb.Circuit = expanded_netlist.circuit_by_name(top_cell_name)

        if not blackbox_devices:
            for d in top_circuit.each_device():
                name = d.name or d.expanded_name()
                info(f"Removing whiteboxed device {name}")
                top_circuit.remove_device(d)

        # create capacitor class
        cap = kdb.DeviceClassCapacitor()
        cap.name = 'PEX_CAP'
        cap.description = "Extracted by kpex/FasterCap PEX"
        expanded_netlist.add(cap)

        fc_gnd_net = top_circuit.create_net('FC_GND')  # create GROUND net
        vsubs_net = top_circuit.create_net("VSUBS")
        nets: List[kdb.Net] = []

        # build table: name -> net
        name2net: Dict[str, kdb.Net] = {n.expanded_name(): n for n in top_circuit.each_net()}

        # find nets for the matrix axes
        pattern = re.compile(r'^g\d+_(.*)$')
        for idx, nn in enumerate(cap_matrix.conductor_names):
            m = pattern.match(nn)
            nn = m.group(1)
            if nn not in name2net:
                raise Exception(f"No net found with name {nn}, net names are: {list(name2net.keys())}")
            n = name2net[nn]
            nets.append(n)

        cap_threshold = 0.0

        def add_parasitic_cap(i: int,
                              j: int,
                              net1: kdb.Net,
                              net2: kdb.Net,
                              cap_value: float):
            if cap_value > cap_threshold:
                c: kdb.Device = top_circuit.create_device(cap, f"Cext_{i}_{j}")
                c.connect_terminal('A', net1)
                c.connect_terminal('B', net2)
                c.set_parameter('C', cap_value)
                if net1 == net2:
                    raise Exception(f"Invalid attempt to create cap {c.name} between "
                                    f"same net {net1} with value {'%.12g' % cap_value}")
            else:
                warning(f"Ignoring capacitance matrix cell [{i},{j}], "
                        f"{'%.12g' % cap_value} is below threshold {'%.12g' % cap_threshold}")

        # -------------------------------------------------------------
        # Example capacitance matrix:
        #     [C11+C12+C13           -C12            -C13]
        #     [-C21           C21+C22+C23            -C23]
        #     [-C31                  -C32     C31+C32+C33]
        # -------------------------------------------------------------
        #
        # - Diagonal elements m[i][i] contain the capacitance over GND (Cii),
        #   but in a sum including all the other values of the row
        #
        # https://www.fastfieldsolvers.com/Papers/The_Maxwell_Capacitance_Matrix_WP110301_R03.pdf
        #
        for i in range(0, cap_matrix.dimension):
            row = cap_matrix[i]
            cap_ii = row[i]
            for j in range(0, cap_matrix.dimension):
                if i == j:
                    continue
                cap_value = -row[j]  # off-diagonals are always stored as negative values
                cap_ii -= cap_value  # subtract summands to filter out Cii
                if j > i:
                    add_parasitic_cap(i=i, j=j,
                                      net1=nets[i], net2=nets[j],
                                      cap_value=cap_value)
            if i > 0:
                add_parasitic_cap(i=i, j=i,
                                  net1=nets[i], net2=nets[0],
                                  cap_value=cap_ii)

        # Short VSUBS and FC_GND together
        #   VSUBS ... substrate block
        #   FC_GND ... FasterCap's GND, i.e. the diagonal Cii elements
        # create capacitor class

        res = kdb.DeviceClassResistor()
        res.name = 'PEX_RES'
        res.description = "Extracted by kpex/FasterCap PEX"
        expanded_netlist.add(res)

        gnd_net = name2net.get('GND', None)
        if not gnd_net:
            gnd_net = top_circuit.create_net('GND')  # create GROUND net

        c: kdb.Device = top_circuit.create_device(res, f"Rext_FC_GND_GND")
        c.connect_terminal('A', fc_gnd_net)
        c.connect_terminal('B', gnd_net)
        c.set_parameter('R', 0)

        c: kdb.Device = top_circuit.create_device(res, f"Rext_VSUBS_GND")
        c.connect_terminal('A', vsubs_net)
        c.connect_terminal('B', gnd_net)
        c.set_parameter('R', 0)

        return expanded_netlist
